#distutils: language = c
#cython: language_level = 3

RESERVED_NCS = 'reserved for NCS compatibility'
RFC_4122 = 'specified in RFC 4122'
RESERVED_MICROSOFT = 'reserved for Microsoft compatibility'
RESERVED_FUTURE = 'reserved for future definition'

from uuid import UUID as _UUID

cdef dict params = {'hex':None, 'bytes': None, 'bytes_le': None, 'fields': None, 'int': None, 'version':None, 'uuid': None}

cdef class UUID:
    def __cinit__(self, hex=None, **params):
        cdef dict data = {}
        data.setdefault('hex', hex)
        data.setdefault('bytes', None)
        data.setdefault('bytes_le', None)
        data.setdefault('fields', None)
        data.setdefault('int', None)
        data.setdefault('version', None)
        data.setdefault('uuid', None)
        data.update(params)

        (hex, _bytes, bytes_le, fields, _int, version, uuid) = data.values()

        if [uuid, hex, _bytes, bytes_le, fields, _int].count(None) != 5:
            raise TypeError('one of the hex, bytes, bytes_le, fields, or int arguments must be given')


        if uuid is not None:
            self._init_uuid(uuid)

        if hex is not None:
            self._init_hex(hex)

        if _bytes is not None:
            self._init_bytes(_bytes)

        if bytes_le is not None:
            self._init_bytes_le(bytes_le)

        if fields is not None:
            self._init_fields(fields)

        if _int is not None:
            self._init_int(_int)

        if version is not None:
            if not 1 <= version <= 5:
                raise ValueError('illegal version number')
            self.value &= ~(0xc000 << 48)
            self.value |= 0x8000 << 48
            self.value &= ~(0xf000 << 64)
            self.value |= version << 76

    cdef _init_fields(self, tuple fields):
        if len(fields) != 6:
            raise ValueError('fields is not a 6-tuple')
        cdef int128 time_low = fields[0]
        cdef int128 time_mid = fields[1]
        cdef int128 time_hi_version = fields[2]
        cdef int128 clock_seq_hi_variant = fields[3]
        cdef int128 clock_seq_low= fields[4]
        cdef int128 node = fields[5]

        if not 0 <= time_low < 1<<32:
            raise ValueError('field 1 out of range (need a 32-bit value)')
        if not 0 <= time_mid < 1<<16:
            raise ValueError('field 2 out of range (need a 16-bit value)')
        if not 0 <= time_hi_version < 1<<16:
            raise ValueError('field 3 out of range (need a 16-bit value)')
        if not 0 <= clock_seq_hi_variant < 1<<8:
            raise ValueError('field 4 out of range (need an 8-bit value)')
        if not 0 <= clock_seq_low < 1<<8:
            raise ValueError('field 5 out of range (need an 8-bit value)')
        if not 0 <= node < 1<<48:
            raise ValueError('field 6 out of range (need a 48-bit value)')
        clock_seq = (clock_seq_hi_variant << 8) | clock_seq_low
        self.value = ((time_low << 96) | (time_mid << 80) | (time_hi_version << 64) | (clock_seq << 48) | node)

    cdef _init_bytes_le(self, bytes bytes_le):
            cdef bytes _bytes = (bytes_le[4-1::-1] + bytes_le[6-1:4-1:-1] + bytes_le[8-1:6-1:-1] + bytes_le[8:])
            self._init_bytes(_bytes)

    cdef _init_bytes(self, bytes _bytes):
        if len(_bytes) != 16:
            raise ValueError('bytes is not a 16-char string')
        self.value = int.from_bytes(_bytes, 'big')

    cdef _init_hex(self, str hex):
        hex= hex.replace('urn:', '').replace('uuid:', '')
        hex = hex.strip('{}').replace('-', '')
        if len(hex) != 32:
            raise ValueError('badly formed hexadecimal UUID string')
        self.value = int(hex, 16)

    cdef _init_int(self, int128 _int):
        self.value = _int

    cdef _init_uuid(self, uuid):
        self.value = int(uuid)

    def __int__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, UUID):
            return self.value == other.value
        elif isinstance(other, _UUID):
            return self.value == int(other)
        else:
            return NotImplemented

    def __lt__(self, other):
        if isinstance(other, UUID):
            return self.value < other.value
        elif isinstance(other, _UUID):
            return self.value < int(other)
        else:
            return NotImplemented

    def __gt__(self, other):
        if isinstance(other, UUID):
            return self.value > other.value
        elif isinstance(other, _UUID):
            return self.value > int(other)
        else:
            return NotImplemented

    def __le__(self, other):
        if isinstance(other, UUID):
            return self.value <= other.value
        elif isinstance(other, _UUID):
            return self.value <= int(other)
        else:
            return NotImplemented

    def __ge__(self, other):
        if isinstance(other, UUID):
            return self.value >= other.value
        elif isinstance(other, _UUID):
            return self.value >= int(other)
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, str(self))

    def __setattr__(self, name, value):
        raise TypeError('UUID objects are immutable')

    def __str__(self):
        hex = '%032x' % self.value
        return '%s-%s-%s-%s-%s' % (
            hex[:8], hex[8:12], hex[12:16], hex[16:20], hex[20:])

    @property
    def bytes(self):
        return self.value.to_bytes(16, 'big')

    @property
    def bytes_le(self):
        bytes = self.bytes
        return (bytes[4-1::-1] + bytes[6-1:4-1:-1] + bytes[8-1:6-1:-1] +
                bytes[8:])

    @property
    def fields(self):
        return (self.time_low, self.time_mid, self.time_hi_version,
                self.clock_seq_hi_variant, self.clock_seq_low, self.node)

    @property
    def time_low(self):
        return self.value >> 96

    @property
    def time_mid(self):
        return (self.value >> 80) & 0xffff

    @property
    def time_hi_version(self):
        return (self.value >> 64) & 0xffff

    @property
    def clock_seq_hi_variant(self):
        return (self.value >> 56) & 0xff

    @property
    def clock_seq_low(self):
        return (self.value >> 48) & 0xff

    @property
    def time(self):
        return (((self.time_hi_version & 0x0fff) << 48) |
                (self.time_mid << 32) | self.time_low)

    @property
    def clock_seq(self):
        return (((self.clock_seq_hi_variant & 0x3f) << 8) |
                self.clock_seq_low)

    @property
    def node(self):
        return self.value & 0xffffffffffff

    @property
    def hex(self):
        return '%032x' % self.value

    @property
    def urn(self):
        return 'urn:uuid:' + str(self)

    @property
    def variant(self):
       if not self.value & (0x8000 << 48):
           return RESERVED_NCS
       elif not self.value & (0x4000 << 48):
           return RFC_4122
       elif not self.value & (0x2000 << 48):
           return RESERVED_MICROSOFT
       else:
           return RESERVED_FUTURE

    @property
    def version(self):
        # The version bits are only meaningful for RFC 4122 UUIDs.
        if self.variant == RFC_4122:
            return (self.value >> 76) & 0xf



from unittest import TestCase
import unittest
from uuid import UUID as _UUID
from cyuuid import UUID

class TestUUIDCompat(TestCase):
    def setUp(self):
        self._uuid = _UUID("a8098c1a-f86e-11da-bd1a-00112444be1e")
        self._cyuuid = UUID("a8098c1a-f86e-11da-bd1a-00112444be1e")

    def test_int(self):
        self.assertEqual(int(self._cyuuid), int(self._uuid))

    def test_hash(self):
        self.assertEqual(hash(self._cyuuid), hash(self._uuid))

    def test_str(self):
        self.assertEqual(str(self._cyuuid), str(self._uuid))

    def test_bytes(self):
        self.assertEqual(self._cyuuid.bytes, self._uuid.bytes)

    def test_bytes_le(self):
        self.assertEqual(self._cyuuid.bytes_le, self._uuid.bytes_le)

    def test_fields(self):
        self.assertEqual(self._cyuuid.fields, self._uuid.fields)

    def test_time_low(self):
        self.assertEqual(self._cyuuid.time_low, self._uuid.time_low)

    def test_time_mid(self):
        self.assertEqual(self._cyuuid.time_mid, self._uuid.time_mid)

    def test_time_hi_version(self):
        self.assertEqual(self._cyuuid.time_hi_version, self._uuid.time_hi_version)

    def test_clock_seq_hi_variant(self):
        self.assertEqual(self._cyuuid.clock_seq_hi_variant, self._uuid.clock_seq_hi_variant)

    def test_clock_seq_low(self):
        self.assertEqual(self._cyuuid.clock_seq_low, self._uuid.clock_seq_low)

    def test_time(self):
        self.assertEqual(self._cyuuid.time, self._uuid.time)

    def test_clock_seq(self):
        self.assertEqual(self._cyuuid.clock_seq, self._uuid.clock_seq)

    def test_node(self):
        self.assertEqual(self._cyuuid.node, self._uuid.node)

    def test_hex(self):
        self.assertEqual(self._cyuuid.hex, self._uuid.hex)

    def test_urn(self):
        self.assertEqual(self._cyuuid.urn, self._uuid.urn)

    def test_variant(self):
        self.assertEqual(self._cyuuid.variant, self._uuid.variant)

    def test_version(self):
        self.assertEqual(self._cyuuid.version, self._uuid.version)




import itertools
import unittest
from uuid import UUID as _UUID
from cyuuid import UUID

class TestUUIDCompatInit(unittest.TestCase):
    def test_int(self):
        _int=223359875637754765292326297443183672862
        uuid = _UUID(int=_int)
        cyuuid = UUID(int=_int)
        self.assertEqual(uuid, cyuuid)

    def test_hex(self):
        hex = "a8098c1a-f86e-11da-bd1a-00112444be1e"
        uuid = _UUID(hex=hex)
        cyuuid = UUID(hex=hex)
        self.assertEqual(uuid, cyuuid)

    def test_bytes(self):
        _bytes = b'\xa8\t\x8c\x1a\xf8n\x11\xda\xbd\x1a\x00\x11$D\xbe\x1e'
        uuid = _UUID(bytes=_bytes)
        cyuuid = UUID(bytes=_bytes)
        self.assertEqual(uuid, cyuuid)

    def test_bytes_le(self):
        bytes_le = b'\x1a\x8c\t\xa8n\xf8\xda\x11\xbd\x1a\x00\x11$D\xbe\x1e'
        uuid = _UUID(bytes_le=bytes_le)
        cyuuid = UUID(bytes_le=bytes_le)
        self.assertEqual(uuid, cyuuid)

    def test_fields(self):
        fields=(0x12345678, 0x1234, 0x5678, 0x12, 0x34, 0x567812345678)
        uuid = _UUID(fields=fields)
        cyuuid = UUID(fields=fields)
        self.assertEqual(uuid, cyuuid)

    def test_version(self):
        hex = "a8098c1a-f86e-11da-bd1a-00112444be1e"
        for version in range(1, 5):
            with self.subTest(version=version):
                uuid = _UUID(hex=hex, version=version)
                cyuuid = UUID(hex=hex, version=version)
                self.assertEqual(uuid, cyuuid)

    def test_multiple_inputs(self):
        hex = "a8098c1a-f86e-11da-bd1a-00112444be1e"
        fields=(0x12345678, 0x1234, 0x5678, 0x12, 0x34, 0x567812345678)
        bytes_le = b'\x1a\x8c\t\xa8n\xf8\xda\x11\xbd\x1a\x00\x11$D\xbe\x1e'
        _bytes = b'\xa8\t\x8c\x1a\xf8n\x11\xda\xbd\x1a\x00\x11$D\xbe\x1e'
        _int=223359875637754765292326297443183672862

        for x in [
                (hex, fields, bytes_le, _bytes, _int),
                (hex, fields, bytes_le, _bytes, None),
                (hex, fields, bytes_le, None, None),
                (hex, fields, None, None, None),
                (hex, None, None, None, None),
                (None, fields, bytes_le, _bytes, _int),
                (hex, None, bytes_le, _bytes, _int),
                (hex, fields, None, _bytes, _int),
                (hex, fields, bytes_le, None, _int),
                (hex, fields, bytes_le, _bytes, None),
                (None, None, None, _bytes, _int),
            ]:

            with self.subTest(params=x):
                msg = "one of the hex, bytes, bytes_le, fields, or int arguments must be given"
                with self.assertRaisesRegex(TypeError, msg):
                    UUID(hex=x[0], bytes=[1], bytes_le=x[2], fields=x[3], int=x[4])



from unittest import TestCase
import unittest
from uuid import UUID as _UUID
from cyuuid import UUID, RESERVED_FUTURE, RFC_4122, RESERVED_NCS, RESERVED_MICROSOFT

class TestUUID(TestCase):
    def setUp(self):
        self._uuid = _UUID("a8098c1a-f86e-11da-bd1a-00112444be1e")

    def test_from_uuid(self):
        self.assertEqual(UUID(uuid=self._uuid), self._uuid)

    def test_not_equal(self):
        self.assertNotEqual(UUID('b8098c1a-f86e-11da-bd1a-00112444be1e'), self._uuid)

    def test_lt(self):
        self.assertLess(UUID('a7098c1a-f86e-11da-bd1a-00112444be1e'), self._uuid)

    def test_gt(self):
        self.assertGreater(UUID('a9098c1a-f86e-11da-bd1a-00112444be1e'), self._uuid)

    def test_lte(self):
        self.assertLessEqual(UUID('a8098c1a-f86e-11da-bd1a-00112444be1e'), self._uuid)
        self.assertLessEqual(UUID('a7098c1a-f86e-11da-bd1a-00112444be1e'), self._uuid)

    def test_gte(self):
        self.assertGreaterEqual(UUID('a8098c1a-f86e-11da-bd1a-00112444be1e'), self._uuid)
        self.assertGreaterEqual(UUID('a9098c1a-f86e-11da-bd1a-00112444be1e'), self._uuid)

    def test_hash(self):
        self.assertEqual(hash(UUID('a8098c1a-f86e-11da-bd1a-00112444be1e')), 2118487228160494877)

    def test_int(self):
        self.assertEqual(int(UUID('a8098c1a-f86e-11da-bd1a-00112444be1e')), 223359875637754765292326297443183672862)

    def test_str(self):
        self.assertEqual(str(UUID('a8098c1a-f86e-11da-bd1a-00112444be1e')), "a8098c1a-f86e-11da-bd1a-00112444be1e")

    def test_hex(self):
        self.assertEqual(UUID('a8098c1a-f86e-11da-bd1a-00112444be1e').hex, "a8098c1af86e11dabd1a00112444be1e")

    def test_variant_NSC(self):
        self.assertEqual(UUID('a8098c1a-f86e-11da-0d1a-00112444be1e').variant, RESERVED_NCS)

    def test_variant_MS(self):
        self.assertEqual(UUID('a8098c1a-f86e-11da-dd1a-00112444be1e').variant, RESERVED_MICROSOFT)

    def test_variant_RFC_4122(self):
        self.assertEqual(UUID('a8098c1a-f86e-11da-bd1a-00112444be1e').variant, RFC_4122)

    def test_variant_FUTURE(self):
        self.assertEqual(UUID('a8098c1a-f86e-11da-ed1a-00112444be1e').variant, RESERVED_FUTURE)

    def test_version(self):
        self.assertEqual(UUID('a8098c1a-f86e-11da-bd1a-00112444be1e').version, 1)



from setuptools import setup, Extension, _install_setup_requires

REQUIRES = [
    'setuptools>=18.0',
    'cython>=0.29'
]

_install_setup_requires({'setup_requires':REQUIRES})

from Cython.Build import cythonize
    
setup(
    name='cyuuid',
    version='0.1.1',
    author='martin.asell',
    description='Cython implementation of RFC4122',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=REQUIRES,
    setup_requires=REQUIRES,
    license='PSF',
    url='https://github.com/masell/cyuuid/',
    packages=['cyuuid', ],
    ext_modules = cythonize(
	Extension('cyuuid.*', sources=["cyuuid/*.pyx"])
    ),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Python Software Foundation License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Cython',
        'Intended Audience :: Developers',
    ]
)



#distutils: language = c
#cython: language_level = 3

cdef extern from "header_int128.h":
    ctypedef unsigned long long int128

cdef class UUID:
    cdef int128 value
    cdef _init_fields(self, tuple fields)
    cdef _init_bytes_le(self, bytes bytes_le)
    cdef _init_bytes(self, bytes _bytes)
    cdef _init_hex(self, str hex)
    cdef _init_int(self, int128 _int)
    cdef _init_uuid(self, uuid)
